# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

__all__ = ['BistrBuilder']

from typing import Iterable, List, Match, Optional

from ._alignment import Alignment
from ._bistr import bistr, String
from ._regex import compile_regex, expand_template
from ._typing import Bounds, Regex, Replacement


class BistrBuilder:
    r"""
    Bidirectionally transformed string builer.

    A `BistrBuilder` builds a transformed version of a source string iteratively.  Each builder has an immutable
    original string, a current string, and the in-progress modified string, with alignments between each.  For example:

    .. code-block:: text

        original: |The| |quick,| |brown| |🦊| |jumps| |over| |the| |lazy| |🐶|
                  |   | |      | |     | |  \ \     \ \    \ \   \ \    \ \   \
        current:  |The| |quick,| |brown| |fox| |jumps| |over| |the| |lazy| |dog|
                  |   | |      / /     /
        modified: |the| |quick| |brown| ...

    The modified string is built in pieces by calling :meth:`replace` to change `n` characters of the current string
    into new ones in the modified string.  Convenience methods like :meth:`skip`, :meth:`insert`, and :meth:`discard`
    are implemented on top of this basic primitive.

        >>> b = BistrBuilder('The quick, brown 🦊 jumps over the lazy 🐶')
        >>> b.skip(17)
        >>> b.peek(1)
        '🦊'
        >>> b.replace(1, 'fox')
        >>> b.skip(21)
        >>> b.peek(1)
        '🐶'
        >>> b.replace(1, 'dog')
        >>> b.is_complete
        True
        >>> b.rewind()
        >>> b.peek(3)
        'The'
        >>> b.replace(3, 'the')
        >>> b.skip(1)
        >>> b.peek(6)
        'quick,'
        >>> b.replace(6, 'quick')
        >>> b.skip_rest()
        >>> s = b.build()
        >>> s.modified
        'the quick brown fox jumps over the lazy dog'
    """

    _original: bistr
    _modified: List[str]
    _alignment: List[Bounds]
    _opos: int
    _mpos: int

    def __init__(self, original: String):
        """
        :param original:
            The string to start from.
        """

        self._original = bistr(original)
        self._modified = []
        self._alignment = [(0, 0)]
        self._opos = 0
        self._mpos = 0

    @property
    def original(self) -> str:
        """
        The original string being modified.
        """
        return self._original.original

    @property
    def current(self) -> str:
        """
        The current string before modifications.
        """
        return self._original.modified

    @property
    def modified(self) -> str:
        """
        The modified string as built so far.
        """
        return ''.join(self._modified)

    @property
    def alignment(self) -> Alignment:
        """
        The alignment built so far from self.current to self.modified.
        """
        return Alignment(self._alignment)

    @property
    def position(self) -> int:
        """
        The position of the builder in self.current.
        """
        return self._opos

    @property
    def remaining(self) -> int:
        """
        The number of characters of the current string left to process.
        """
        return len(self.current) - self._opos

    @property
    def is_complete(self) -> bool:
        """
        Whether we've completely processed the string.  In other words, whether the modified string aligns with the end
        of the current string.
        """
        return self.remaining == 0

    def peek(self, n: int) -> str:
        """
        Peek at the next `n` characters of the original string.
        """
        return self.current[self._opos:self._opos+n]

    def _advance(self, ocount: int, mcount: int) -> None:
        self._opos += ocount
        self._mpos += mcount
        if ocount > 0 or mcount > 0:
            self._alignment.append((self._opos, self._mpos))

    def skip(self, n: int) -> None:
        """
        Skip the next `n` characters, copying them unchanged.
        """
        if n > 0:
            self._modified.append(self.peek(n))
            for i in range(n):
                self._advance(1, 1)

    def skip_rest(self) -> None:
        """
        Skip the rest of the string, copying it unchanged.
        """
        self.skip(self.remaining)

    def insert(self, string: str) -> None:
        """
        Insert a substring into the string.
        """
        self.replace(0, string)

    def discard(self, n: int) -> None:
        """
        Discard a portion of the original string.
        """
        self.replace(n, '')

    def discard_rest(self) -> None:
        """
        Discard the rest of the original string.
        """
        self.discard(self.remaining)

    def replace(self, n: int, repl: str) -> None:
        """
        Replace the next `n` characters with a new string.
        """
        if len(repl) > 0:
            self._modified.append(repl)
        self._advance(n, len(repl))

    def append(self, bs: bistr) -> None:
        """
        Append a bistr.  The original value of the bistr must match the current string being processed.
        """
        if bs.original != self.peek(len(bs.original)):
            raise ValueError("bistr doesn't match the current string")
        self._modified.append(bs.modified)
        for (o0, m0), (o1, m1) in zip(bs.alignment, bs.alignment[1:]):
            self._advance(o1 - o0, m1 - m0)

    def _match(self, regex: Regex) -> Optional[Match[str]]:
        pattern = compile_regex(regex)
        return pattern.match(self.current, pos=self._opos)

    def _search(self, regex: Regex) -> Optional[Match[str]]:
        pattern = compile_regex(regex)
        return pattern.search(self.current, pos=self._opos)

    def _finditer(self, regex: Regex) -> Iterable[Match[str]]:
        pattern = compile_regex(regex)
        return pattern.finditer(self.current, pos=self._opos)

    def skip_match(self, regex: Regex) -> bool:
        """
        Skip a substring matching a regex, copying it unchanged.

        :param regex:
            The (possibly compiled) regular expression to match.
        :returns:
            Whether a match was found.
        """

        match = self._match(regex)
        if match:
            self.skip(match.end() - match.start())
            return True
        else:
            return False

    def discard_match(self, regex: Regex) -> bool:
        """
        Discard a substring that matches a regex.

        :param regex:
            The (possibly compiled) regular expression to match.
        :returns:
            Whether a match was found.
        """

        match = self._match(regex)
        if match:
            self.discard(match.end() - match.start())
            return True
        else:
            return False

    def replace_match(self, regex: Regex, repl: Replacement) -> bool:
        """
        Replace a substring that matches a regex.

        :param regex:
            The (possibly compiled) regular expression to match.
        :param repl:
            The replacement to use.  Can be a string, which is interpreted as in :meth:`re.Match.expand`, or a
            `callable`, which will receive each match and return the replacement string.
        :returns:
            Whether a match was found.
        """

        match = self._match(regex)
        if match:
            self.replace(match.end() - match.start(), expand_template(match, repl))
            return True
        else:
            return False

    def replace_next(self, regex: Regex, repl: Replacement) -> bool:
        """
        Replace the next occurence of a regex.

        :param regex:
            The (possibly compiled) regular expression to match.
        :param repl:
            The replacement to use.
        :returns:
            Whether a match was found.
        """

        match = self._search(regex)
        if match:
            self.skip(match.start() - self._opos)
            self.replace(match.end() - match.start(), expand_template(match, repl))
            return True
        else:
            return False

    def replace_all(self, regex: Regex, repl: Replacement) -> None:
        """
        Replace all occurences of a regex.

        :param regex:
            The (possibly compiled) regular expression to match.
        :param repl:
            The replacement to use.
        """

        for match in self._finditer(regex):
            self.skip(match.start() - self._opos)
            self.replace(match.end() - match.start(), expand_template(match, repl))
        self.skip_rest()

    def build(self) -> bistr:
        """
        Build the `bistr`.

        :returns:
            A `bistr` from the original string to the new modified string.
        :raises:
            :class:`ValueError` if the modified string is not completely built yet.
        """

        if not self.is_complete:
            raise ValueError(f'The string is not completely built yet ({self.remaining} characters remaining)')

        alignment = self._original.alignment.compose(self.alignment)
        return bistr(self.original, self.modified, alignment)

    def rewind(self) -> None:
        """
        Reset this builder to apply another transformation.

        :raises:
            :class:`ValueError` if the modified string is not completely built yet.
        """
        self._original = self.build()
        self._modified = []
        self._alignment = [(0, 0)]
        self._opos = 0
        self._mpos = 0



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from __future__ import annotations

__all__ = ['Alignment']

import bisect
from typing import Any, Callable, Iterable, Iterator, List, Optional, Sequence, Tuple, TypeVar, Union, cast, overload

from ._typing import AnyBounds, Bounds, Index, Range


T = TypeVar('T')
U = TypeVar('U')
Real = Union[int, float]
CostFn = Callable[[Optional[T], Optional[U]], Real]


class Alignment:
    r"""
    An alignment between two related sequences.

    Consider this alignments between two strings:

    .. code-block:: text

        |it's| |aligned!|
        |    \ \        |
        |it is| |aligned|

    An alignment stores all the indices that are known to correspond between the original and modified sequences.  For
    the above example, it would be

        >>> a = Alignment([
        ...     (0, 0),
        ...     (4, 5),
        ...     (5, 6),
        ...     (13, 13),
        ... ])

    Alignments can be used to answer questions like, "what's the smallest range of the original sequence that is
    guaranteed to contain this part of the modified sequence?"  For example, the range ``(0, 5)`` ("it is") is known to
    match the range ``(0, 4)`` ("it's") of the original sequence:

        >>> a.original_bounds(0, 5)
        (0, 4)

    Results may be imprecise if the alignment is too course to match the exact inputs:

        >>> a.original_bounds(0, 2)
        (0, 4)

    A more granular alignment like this:

    .. code-block:: text

        |i|t|'s| |a|l|i|g|n|e|d|!|
        | | |  \ \ \ \ \ \ \ \ \ /
        |i|t| is| |a|l|i|g|n|e|d|

    .. doctest::

        >>> a = Alignment([
        ...     (0, 0), (1, 1), (2, 2), (4, 5), (5, 6), (6, 7), (7, 8),
        ...     (8, 9), (9, 10), (10, 11), (11, 12), (12, 13), (13, 13),
        ... ])

    Can be more precise:

        >>> a.original_bounds(0, 2)
        (0, 2)
    """

    __slots__ = ('_original', '_modified')

    _original: List[int]
    _modified: List[int]

    def __init__(self, values: Iterable[Bounds]):
        """
        :param values:
            The sequence of aligned indices.  Each element should be a tuple ``(x, y)``, where `x` is the original
            sequence position and `y` is the modified sequence position.
        """

        self._original = []
        self._modified = []
        for i, j in values:
            if self._original:
                if i < self._original[-1]:
                    raise ValueError('Original sequence position moved backwards')
                elif j < self._modified[-1]:
                    raise ValueError('Modified sequence position moved backwards')
                elif i == self._original[-1] and j == self._modified[-1]:
                    continue

            self._original.append(i)
            self._modified.append(j)

        if not self._original:
            raise ValueError('No sequence positions to align')

    @classmethod
    def _create(cls, original: List[int], modified: List[int]) -> Alignment:
        result: Alignment = super().__new__(cls)
        result._original = original
        result._modified = modified
        return result

    def __str__(self) -> str:
        i, j = self._original[0], self._original[-1]
        k, l = self._modified[0], self._modified[-1]
        if self._original == list(range(i, j + 1)) and self._modified == list(range(k, l + 1)):
            return f'[{i}:{j}⇋{k}:{l}]'
        else:
            return '[' + ', '.join(f'{i}⇋{j}' for i, j in self) + ']'

    def __repr__(self) -> str:
        i, j = self._original[0], self._original[-1]
        if self._original == list(range(i, j + 1)) and self._modified == list(range(i, j + 1)):
            if i == 0:
                return f'Alignment.identity({j})'
            else:
                return f'Alignment.identity({i}, {j})'
        else:
            return 'Alignment([' + ', '.join(map(repr, self)) + '])'

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Alignment):
            return (self._original, self._modified) == (other._original, other._modified)
        else:
            return NotImplemented

    @classmethod
    def _parse_bounds(cls, args: Tuple[AnyBounds, ...]) -> Bounds:
        l = len(args)
        if l == 0:
            raise TypeError('Not enough arguments')
        elif l == 1:
            arg = args[0]
            if isinstance(arg, range):
                return arg.start, arg.stop
            elif isinstance(arg, slice):
                if arg.start is None or arg.stop is None:
                    raise ValueError('slice with unspecified bounds')
                return arg.start, arg.stop
            elif isinstance(arg, tuple):
                return arg
            else:
                return 0, arg
        elif l == 2:
            return cast(Bounds, args)
        else:
            raise TypeError('Too many arguments')

    @classmethod
    def _parse_optional_bounds(cls, args: Tuple[AnyBounds, ...]) -> Union[Bounds, Tuple[None, None]]:
        if len(args) == 0:
            return None, None
        else:
            return cls._parse_bounds(args)

    @overload
    @classmethod
    def identity(cls, __length: int) -> Alignment: ...

    @overload
    @classmethod
    def identity(cls, __start: int, __stop: int) -> Alignment: ...

    @overload
    @classmethod
    def identity(cls, __bounds: Range) -> Alignment: ...

    @classmethod
    def identity(cls, *args: Union[int, range, slice, Bounds]) -> Alignment:
        """
        Create an identity alignment, which maps all intervals to themselves.  You can pass the size of the sequence:

            >>> Alignment.identity(5)
            Alignment.identity(5)

        or the start and end positions:

            >>> Alignment.identity(1, 5)
            Alignment.identity(1, 5)

        or a range-like object (:class:`range`, :class:`slice`, or ``Tuple[int, int]``):

            >>> Alignment.identity(range(1, 5))
            Alignment.identity(1, 5)
        """

        start, stop = cls._parse_bounds(args)
        values = list(range(start, stop + 1))
        return cls._create(values, values)

    @classmethod
    def _infer_costs(cls, original: Sequence[T], modified: Sequence[U], cost_fn: CostFn[T, U]) -> List[Real]:
        """
        The Needleman–Wunsch or Wagner–Fischer algorithm.  Here we use it in a way that only computes the final row of
        costs, without finding the alignment itself.  Hirschberg's algorithm uses it as a subroutine to find the optimal
        alignment in less than O(N*M) space.

        https://en.wikipedia.org/wiki/Needleman%E2%80%93Wunsch_algorithm
        https://en.wikipedia.org/wiki/Wagner%E2%80%93Fischer_algorithm
        """

        row: List[Real] = [0]
        for i, m in enumerate(modified):
            cost = row[i] + cost_fn(None, m)
            row.append(cost)

        prev: List[Real] = [0] * len(row)

        for o in original:
            prev, row = row, prev
            row[0] = prev[0] + cost_fn(o, None)

            for i, m in enumerate(modified):
                sub_cost = prev[i] + cost_fn(o, m)
                del_cost = prev[i + 1] + cost_fn(o, None)
                ins_cost = row[i] + cost_fn(None, m)
                row[i + 1] = min(sub_cost, del_cost, ins_cost)

        return row

    @classmethod
    def _infer_matrix(cls, original: Sequence[T], modified: Sequence[U], cost_fn: CostFn[T, U]) -> List[Bounds]:
        """
        The Needleman–Wunsch or Wagner–Fischer algorithm, using the entire matrix to compute the optimal alignment.
        """

        row: List[Tuple[Real, int, int]] = [(0, -1, -1)]
        for j, m in enumerate(modified):
            cost = row[j][0] + cost_fn(None, m)
            row.append((cost, 0, j))

        matrix = [row]

        for i, o in enumerate(original):
            prev = matrix[i]
            cost = prev[0][0] + cost_fn(o, None)
            row = [(cost, i, 0)]

            for j, m in enumerate(modified):
                cost = prev[j][0] + cost_fn(o, m)
                x, y = i, j

                del_cost = prev[j + 1][0] + cost_fn(o, None)
                if del_cost < cost:
                    cost = del_cost
                    x, y = i, j + 1

                ins_cost = row[j][0] + cost_fn(None, m)
                if ins_cost < cost:
                    cost = ins_cost
                    x, y = i + 1, j

                row.append((cost, x, y))

            matrix.append(row)

        result = []
        i = len(matrix) - 1
        j = len(matrix[i]) - 1
        while i >= 0:
            result.append((i, j))
            _, i, j = matrix[i][j]

        result.reverse()
        return result

    @classmethod
    def _infer_recursive(cls, original: Sequence[T], modified: Sequence[U], cost_fn: CostFn[T, U]) -> List[Bounds]:
        """
        Hirschberg's algorithm for computing optimal alignments in linear space.

        https://en.wikipedia.org/wiki/Hirschberg's_algorithm
        """

        if len(original) <= 1 or len(modified) <= 1:
            return cls._infer_matrix(original, modified, cost_fn)

        omid = len(original) // 2
        oleft = original[:omid]
        oright = original[omid:]

        lcosts = cls._infer_costs(oleft, modified, cost_fn)
        rcosts = cls._infer_costs(oright[::-1], modified[::-1], cost_fn)[::-1]

        mmid = min(range(len(lcosts)), key=lambda i: lcosts[i] + rcosts[i])
        mleft = modified[:mmid]
        mright = modified[mmid:]

        left = cls._infer_recursive(oleft, mleft, cost_fn)
        right = cls._infer_recursive(oright, mright, cost_fn)
        for (o, m) in right:
            left.append((o + omid, m + mmid))
        return left

    @classmethod
    def infer(cls, original: Sequence[T], modified: Sequence[U], cost_fn: Optional[CostFn[T, U]] = None) -> Alignment:
        """
        Infer the alignment between two sequences with the lowest edit distance.

            >>> Alignment.infer('color', 'color')
            Alignment.identity(5)
            >>> a = Alignment.infer('color', 'colour')
            >>> # 'ou' -> 'o'
            >>> a.original_bounds(3, 5)
            (3, 4)

        Warning: this operation has time complexity ``O(N*M)``, where `N` and `M` are the lengths of the original and
        modified sequences, and so should only be used for relatively short sequences.

        :param original:
            The original sequence.
        :param modified:
            The modified sequence.
        :param cost_fn:
            A function returning the cost of performing an edit.  ``cost_fn(a, b)`` returns the cost of replacing `a`
            with `b`.  ``cost_fn(a, None)`` returns the cost of deleting `a`, and ``cost_fn(None, b)`` returns the cost
            of inserting `b`.  By default, all operations have cost 1 except replacing identical elements, which has
            cost 0.
        :returns:
            The inferred alignment.
        """

        if cost_fn is None:
            real_cost_fn: CostFn[T, U] = lambda a, b: int(a != b)
        else:
            real_cost_fn = cost_fn

        if len(original) < len(modified):
            swapped_cost_fn = lambda a, b: real_cost_fn(b, a)
            result = cls._infer_recursive(modified, original, swapped_cost_fn)
            return Alignment(result).inverse()
        else:
            result = cls._infer_recursive(original, modified, real_cost_fn)
            return Alignment(result)

    def __iter__(self) -> Iterator[Tuple[int, int]]:
        return zip(self._original, self._modified)

    def __len__(self) -> int:
        return len(self._original)

    @overload
    def __getitem__(self, index: int) -> Bounds: ...

    @overload
    def __getitem__(self, index: slice) -> Alignment: ...

    def __getitem__(self, index: Index) -> Union[Bounds, Alignment]:
        """
        Indexing an alignment returns the nth pair of aligned positions:

            >>> a = Alignment.identity(5)
            >>> a[3]
            (3, 3)

        Slicing an alignment returns a new alignment with a subrange of its values:

            >>> a[1:5]
            Alignment.identity(1, 4)
        """

        if isinstance(index, slice):
            start, stop, stride = index.indices(len(self))
            if stride != 1:
                raise ValueError('Non-unit strides not supported')
            return self._create(self._original[index], self._modified[index])
        else:
            return (self._original[index], self._modified[index])

    def shift(self, delta_o: int, delta_m: int) -> Alignment:
        """
        Shift this alignment.

        :param delta_o:
            The distance to shift the original sequence.
        :param delta_m:
            The distance to shift the modified sequence.
        :returns:
            An alignment with all the positions shifted by the given amounts.
        """

        return self._create(
            [o + delta_o for o in self._original],
            [m + delta_m for m in self._modified],
        )

    def _search(self, source: List[int], start: int, stop: int) -> Bounds:
        first = bisect.bisect_right(source, start)
        if first == 0:
            raise IndexError('range start too small')
        first -= 1

        last = bisect.bisect_left(source, stop, first)
        if last == len(source):
            raise IndexError('range end too big')

        return first, last

    def _bounds(self, source: List[int], target: List[int], args: Tuple[AnyBounds, ...]) -> Bounds:
        start, stop = self._parse_optional_bounds(args)
        if start is None or stop is None:
            i, j = 0, -1
        else:
            i, j = self._search(source, start, stop)
        return (target[i], target[j])

    def original_bounds(self, *args: AnyBounds) -> Bounds:
        """
        Maps a subrange of the modified sequence to the original sequence.  Can be called with either two arguments:

            >>> a = Alignment.identity(5).shift(1, 0)
            >>> a.original_bounds(1, 3)
            (2, 4)

        or with a range-like object:

            >>> a.original_bounds(range(1, 3))
            (2, 4)

        With no arguments, returns the bounds of the entire original sequence:

            >>> a.original_bounds()
            (1, 6)

        :returns:
            The corresponding bounds in the original sequence.
        """

        return self._bounds(self._modified, self._original, args)

    def original_range(self, *args: AnyBounds) -> range:
        """
        Like :meth:`original_bounds`, but returns a :class:`range`.
        """
        return range(*self.original_bounds(*args))

    def original_slice(self, *args: AnyBounds) -> slice:
        """
        Like :meth:`original_bounds`, but returns a :class:`slice`.
        """
        return slice(*self.original_bounds(*args))

    def modified_bounds(self, *args: AnyBounds) -> Bounds:
        """
        Maps a subrange of the original sequence to the modified sequence.  Can be called with either two arguments:

            >>> a = Alignment.identity(5).shift(1, 0)
            >>> a.modified_bounds(2, 4)
            (1, 3)

        or with a range-like object:

            >>> a.modified_bounds(range(2, 4))
            (1, 3)

        With no arguments, returns the bounds of the entire modified sequence:

            >>> a.modified_bounds()
            (0, 5)

        :returns:
            The corresponding bounds in the modified sequence.
        """

        return self._bounds(self._original, self._modified, args)

    def modified_range(self, *args: AnyBounds) -> range:
        """
        Like :meth:`modified_bounds`, but returns a :class:`range`.
        """
        return range(*self.modified_bounds(*args))

    def modified_slice(self, *args: AnyBounds) -> slice:
        """
        Like :meth:`modified_bounds`, but returns a :class:`range`.
        """
        return slice(*self.modified_bounds(*args))

    def slice_by_original(self, *args: AnyBounds) -> Alignment:
        """
        Slice this alignment by a span of the original sequence.

            >>> a = Alignment.identity(5).shift(1, 0)
            >>> a.slice_by_original(2, 4)
            Alignment([(2, 1), (3, 2), (4, 3)])

        :returns:
            The slice of this alignment that corresponds with the given span of the original sequence.
        """

        start, stop = self._parse_bounds(args)
        first, last = self._search(self._original, start, stop)
        original = self._original[first:last+1]
        original = [min(max(i, start), stop) for i in original]
        modified = self._modified[first:last+1]
        return self._create(original, modified)

    def slice_by_modified(self, *args: AnyBounds) -> Alignment:
        """
        Slice this alignment by a span of the modified sequence.

            >>> a = Alignment.identity(5).shift(1, 0)
            >>> a.slice_by_modified(1, 3)
            Alignment([(2, 1), (3, 2), (4, 3)])

        :returns:
            The slice of this alignment that corresponds with the given span of the modified sequence.
        """

        start, stop = self._parse_bounds(args)
        first, last = self._search(self._modified, start, stop)
        original = self._original[first:last+1]
        modified = self._modified[first:last+1]
        modified = [min(max(i, start), stop) for i in modified]
        return self._create(original, modified)

    def __add__(self, other: Any) -> Alignment:
        """
        Concatenate two alignments.
        """

        if not isinstance(other, Alignment):
            return NotImplemented

        o_orig = other._original
        o_mod = other._modified

        if o_orig[0] < self._original[-1]:
            raise ValueError('Original sequence position moved backwards')
        elif o_mod[0] < self._modified[-1]:
            raise ValueError('Modified sequence position moved backwards')
        elif o_orig[0] == self._original[-1] and o_mod[0] == self._modified[-1]:
            o_orig = o_orig[1:]
            o_mod = o_mod[1:]

        return self._create(self._original + o_orig, self._modified + o_mod)

    def compose(self, other: Alignment) -> Alignment:
        """
        :returns:
            A new alignment equivalent to applying this one first, then the `other`.
        """

        if self.modified_bounds() != other.original_bounds():
            raise ValueError('Incompatible alignments')

        original = []
        modified = []
        i, i_max = 0, len(self)
        j, j_max = 0, len(other)

        while i < i_max:
            # Map self._original[i] to its lower bound in other
            while self._modified[i] > other._original[j]:
                j += 1
            while self._modified[i] < other._original[j] and self._modified[i + 1] <= other._original[j]:
                i += 1
            original.append(self._original[i])
            modified.append(other._modified[j])

            # Map self._original[i] to its upper bound in other (if it's different)
            while i + 1 < i_max and self._original[i] == self._original[i + 1]:
                i += 1

            needs_upper = False
            while j + 1 < j_max and self._modified[i] >= other._original[j + 1]:
                needs_upper = True
                j += 1
            if needs_upper:
                original.append(self._original[i])
                modified.append(other._modified[j])

            i += 1

        return self._create(original, modified)

    def inverse(self) -> Alignment:
        """
        :returns:
            The inverse of this alignment, from the modified to the original sequence.
        """
        return self._create(self._modified, self._original)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from bistring import bistr, Alignment, BistrBuilder


def test_chunk_words():
    builder = BistrBuilder('  the quick  brown fox ')
    builder.discard(2)
    builder.replace(3, 'the')
    builder.skip(1)
    builder.replace(5, 'quick')
    builder.replace(2, ' ')
    builder.replace(5, 'brown')
    builder.skip(1)
    builder.replace(3, 'fox')
    builder.discard(1)
    bs = builder.build()

    assert bs.original == '  the quick  brown fox '
    assert bs.modified == 'the quick brown fox'

    assert bs[0:1].original == 'the'
    assert bs[1:2].original == 'the'
    assert bs[2:3].original == 'the'

    assert bs[0:3].original == 'the'
    assert bs[1:3].original == 'the'

    assert bs[0:4].original == 'the '
    assert bs[1:4].original == 'the '

    assert bs[3:4].original == ' '
    assert bs[9:10].original == '  '

    assert bs[4:15].original == 'quick  brown'
    assert bs[5:14].original == 'quick  brown'

    assert bs[0:0].original == ''
    assert bs[10:10].original == ''


def test_chunk_chars():
    builder = BistrBuilder('  the quick  brown fox ')
    builder.discard_match(r'\s+')
    while not builder.is_complete:
        builder.skip_match(r'\S+')
        builder.replace_match(r'\s+(?=\S)', ' ')
        builder.discard_match(r'\s+$')

    bs = builder.build()

    assert bs.original == '  the quick  brown fox '
    assert bs.modified == 'the quick brown fox'

    assert bs[0:1].original == 't'
    assert bs[1:2].original == 'h'
    assert bs[2:3].original == 'e'

    assert bs[0:3].original == 'the'
    assert bs[1:3].original == 'he'

    assert bs[0:4].original == 'the '
    assert bs[1:4].original == 'he '

    assert bs[3:4].original == ' '
    assert bs[9:10].original == '  '

    assert bs[4:15].original == 'quick  brown'
    assert bs[5:14].original == 'uick  brow'

    assert bs[0:0].original == ''
    assert bs[10:10].original == ''


def test_empty_string():
    builder = BistrBuilder('')
    bs = builder.build()
    assert bs.original == ''
    assert bs.modified == ''
    assert bs[0:0].original == ''


def test_iterative():
    builder = BistrBuilder("I wish I wouldn't've spent one thousand dollars.")
    builder.skip_match(r'[^.]*')
    builder.discard_rest()
    builder.rewind()
    builder.skip_match(r'I wish I ');
    builder.replace_match(r"wouldn't've", 'would not have');
    builder.skip_match(r' spent ');
    builder.replace_match(r'one thousand dollars', '$1,000');

    bs = builder.build()
    assert bs.original == "I wish I wouldn't've spent one thousand dollars."
    assert bs.modified == 'I wish I would not have spent $1,000'


def test_replace_matches():
    builder = BistrBuilder('the cheese that the mouse that the cat that the dog chased played with ate')
    builder.replace_next(r'that', 'which')
    builder.replace_all(r'that', 'whom')

    bs = builder.build()
    assert bs.original == 'the cheese that the mouse that the cat that the dog chased played with ate'
    assert bs.modified == 'the cheese which the mouse whom the cat whom the dog chased played with ate'


def test_replace_backreference():
    builder = BistrBuilder("it doesn't work and stuff doesn't get replaced")
    builder.replace_all(r"\bdoesn't (\S+)", r'\1s')

    bs = builder.build()
    assert bs.original == "it doesn't work and stuff doesn't get replaced"
    assert bs.modified == 'it works and stuff gets replaced'


def test_append():
    builder = BistrBuilder('hello WORLD')
    builder.append(bistr(builder.peek(5)).upper('en_US'))
    builder.skip(1)
    builder.append(bistr(builder.peek(5)).lower('en_US'))

    bs = builder.build()
    assert bs[1:4] == bistr('ell', 'ELL', Alignment.identity(3))
    assert bs[7:10] == bistr('ORL', 'orl', Alignment.identity(3))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from bistring import Alignment, bistr
import pytest
import unicodedata


def test_new():
    pytest.raises(TypeError, bistr, 42)
    pytest.raises(TypeError, bistr, 'fourty-two', 42)
    pytest.raises(TypeError, bistr, 'fourty-two', '42', 42)

    pytest.raises(ValueError, bistr, 'fourty-two', '42', Alignment([
        (0, 0),
        (9, 2),
    ]))
    pytest.raises(ValueError, bistr, 'fourty-two', '42', Alignment([
        (0, 0),
        (10, 1),
    ]))

    bistr('42')
    bistr('fourty-two', '42')
    bistr('fourty-two', '42', Alignment([
        (0, 0),
        (6, 1),
        (7, 1),
        (10, 2),
    ]))


def test_infer():
    bs = bistr.infer('test', 'test')
    assert bs == bistr('test', 'test', Alignment.identity(4))

    bs = bistr.infer('color', 'colour')
    assert bs[3:5].original == 'o'

    assert bs.inverse() == bistr.infer('colour', 'color')

    bs = bistr.infer("--Hello, world!--", "hello world")
    assert bs[:5] == bistr("Hello", "hello", Alignment.identity(5))
    assert bs[6:] == bistr("world")

    bs = bistr.infer(
        '🅃🄷🄴 🅀🅄🄸🄲🄺, 🄱🅁🄾🅆🄽 🦊 🄹🅄🄼🄿🅂 🄾🅅🄴🅁 🅃🄷🄴 🄻🄰🅉🅈 🐶',
        'the quick brown fox jumps over the lazy dog',
    )
    assert bs[0:3] == bistr('🅃🄷🄴', 'the', Alignment.identity(3))
    assert bs[4:9] == bistr('🅀🅄🄸🄲🄺', 'quick', Alignment.identity(5))
    assert bs[10:15] == bistr('🄱🅁🄾🅆🄽', 'brown', Alignment.identity(5))
    assert bs[16:19].original == '🦊'
    assert bs[16:19].modified == 'fox'
    assert bs[20:25] == bistr('🄹🅄🄼🄿🅂', 'jumps', Alignment.identity(5))
    assert bs[40:43].original == '🐶'
    assert bs[40:43].modified == 'dog'

    bs = bistr.infer(
        'Ṫḧë qüïċḳ, ḅṛöẅṅ 🦊 jüṁṗṡ öṿëṛ ẗḧë ḷäżÿ 🐶',
        'the quick brown fox jumps over the lazy dog',
    )
    assert bs[0:3] == bistr('Ṫḧë', 'the', Alignment.identity(3))
    assert bs[4:9] == bistr('qüïċḳ', 'quick', Alignment.identity(5))
    assert bs[10:15] == bistr('ḅṛöẅṅ', 'brown', Alignment.identity(5))
    assert bs[16:19].original == '🦊'
    assert bs[16:19].modified == 'fox'
    assert bs[20:25] == bistr('jüṁṗṡ', 'jumps', Alignment.identity(5))
    assert bs[40:43].original == '🐶'
    assert bs[40:43].modified == 'dog'

    bs = bistr.infer('Z̴̡̪̫̖̥̔̿̃̈̏̎͠͝á̸̪̠̖̻̬̖̪̞͙͇̮̠͎̆͋́̐͌̒͆̓l̶͉̭̳̤̬̮̩͎̟̯̜͇̥̠̘͑͐̌͂̄́̀̂̌̈͛̊̄̚͜ģ̸̬̼̞̙͇͕͎̌̾̒̐̿̎̆̿̌̃̏̌́̾̈͘͜o̶̢̭͕͔̩͐ ̴̡̡̜̥̗͔̘̦͉̣̲͚͙̐̈́t̵͈̰̉̀͒̎̈̿̔̄̽͑͝͠ẹ̵̫̲̫̄͜͜x̵͕̳͈̝̤̭̼̼̻͓̿̌̽̂̆̀̀̍̒͐́̈̀̚͝t̸̡̨̥̺̣̟͎̝̬̘̪͔͆́̄̅̚', 'Zalgo text')
    for i, c in enumerate(bs):
        assert bs[i:i+1].original.startswith(c)


def test_concat():
    bs = bistr('  ', '')
    bs += 'Hello'
    bs += bistr('  ', ' ')
    bs += 'world!'
    bs += bistr('  ', '')

    assert bs.original == '  Hello  world!  '
    assert bs.modified == 'Hello world!'

    bs = bs[4:7]
    assert bs.original == 'o  w'
    assert bs.modified == 'o w'

    bs = bs[1:2]
    assert bs.original == '  '
    assert bs.modified == ' '


def test_find_index():
    bs = bistr('dysfunction')

    assert bs.find('dis') == -1
    assert bs.find('fun') == 3
    assert bs.find('n') == 5
    assert bs.find('n', 6) == 10

    assert bs.find_bounds('dis') == (-1, -1)
    assert bs.find_bounds('fun') == (3, 6)
    assert bs.find_bounds('n') == (5, 6)
    assert bs.find_bounds('n', 6) == (10, 11)

    pytest.raises(ValueError, bs.index, 'dis')
    pytest.raises(ValueError, bs.index_bounds, 'dis')

    assert bs.index('fun') == 3
    assert bs.index_bounds('fun') == (3, 6)
    assert bs.index_bounds('n') == (5, 6)
    assert bs.index_bounds('n', 6) == (10, 11)


def test_rfind_rindex():
    bs = bistr('dysfunction')

    assert bs.rfind('dis') == -1
    assert bs.rfind('fun') == 3
    assert bs.rfind('n') == 10
    assert bs.rfind('n', None, 9) == 5

    assert bs.rfind_bounds('dis') == (-1, -1)
    assert bs.rfind_bounds('fun') == (3, 6)
    assert bs.rfind_bounds('n') == (10, 11)
    assert bs.rfind_bounds('n', None, 9) == (5, 6)

    pytest.raises(ValueError, bs.index, 'dis')
    pytest.raises(ValueError, bs.index_bounds, 'dis')

    assert bs.rindex('fun') == 3
    assert bs.rindex_bounds('fun') == (3, 6)
    assert bs.rindex_bounds('n') == (10, 11)
    assert bs.rindex_bounds('n', None, 9) == (5, 6)


def test_starts_ends_with():
    bs = bistr('Beginning, middle, ending')

    assert bs.startswith('Begin')
    assert bs.endswith('ing')

    assert not bs.startswith('ending')
    assert not bs.endswith('Beginning')

    assert bs.startswith(('Begin', 'End'))
    assert bs.endswith(('beginning', 'ending'))


def test_justify():
    bs = bistr('Hello world!')

    assert bs.center(5) == bs
    assert bs.center(20) == bistr('', '    ') + bs + bistr('', '    ')
    assert bs.center(21) == bistr('', '    ') + bs + bistr('', '     ')

    assert bs.ljust(5) == bs
    assert bs.ljust(16) == bs + bistr('', '    ')

    assert bs.rjust(5) == bs
    assert bs.rjust(16) == bistr('', '    ') + bs


def test_split():
    bs = bistr('1,2,3')
    assert bs.split(',') == [bistr('1'), bistr('2'), bistr('3')]
    assert bs.split(',', 1) == [bistr('1'), bistr('2,3')]

    assert bistr('1,2,,3,').split(',') == [bistr('1'), bistr('2'), bistr(''), bistr('3'), bistr('')]

    assert bistr('').split(',') == [bistr('')]

    assert bistr('1<>2<>3').split('<>') == [bistr('1'), bistr('2'), bistr('3')]

    bs = bistr('   1   2   3   ')
    assert bs.split() == [bistr('1'), bistr('2'), bistr('3')]
    assert bs.split(maxsplit=-1) == [bistr('1'), bistr('2'), bistr('3')]
    assert bs.split(maxsplit=2) == [bistr('1'), bistr('2'), bistr('3   ')]
    assert bs.split(maxsplit=1) == [bistr('1'), bistr('2   3   ')]

    assert bistr('').split() == []


def test_partition():
    bs = bistr('left::middle::right')

    left, sep, right = bs.partition('::')
    assert left == bistr('left')
    assert sep == bistr('::')
    assert right == bistr('middle::right')

    left, sep, right = bs.partition(':::')
    assert left == bs
    assert sep == bistr('')
    assert right == bistr('')

    left, sep, right = bs.rpartition('::')
    assert left == bistr('left::middle')
    assert sep == bistr('::')
    assert right == bistr('right')

    left, sep, right = bs.rpartition(':::')
    assert left == bistr('')
    assert sep == bistr('')
    assert right == bs


def test_expandtabs():
    bs = bistr(' \tHello\t\tworld!\n\tGoodbye \tworld!')
    bs = bs.expandtabs()

    assert bs.modified == bs.original.expandtabs()
    assert bs[0:1] == bistr(' ')
    assert bs[1:8] == bistr('\t', '       ')
    assert bs[8:13] == bistr('Hello')
    assert bs[13:16] == bistr('\t', '   ')
    assert bs[16:24] == bistr('\t', '        ')
    assert bs[24:30] == bistr('world!')
    assert bs[30:31] == bistr('\n')


def test_strip():
    bs = bistr('  Hello  world!  ')
    assert bs.original == '  Hello  world!  '
    assert bs.modified == '  Hello  world!  '

    bs = bs.strip()
    assert bs.original == '  Hello  world!  '
    assert bs.modified == 'Hello  world!'

    bs = bistr('    ').strip()
    assert bs.modified == ''
    assert bs.original == '    '


def test_casefold():
    # 'Híﬃ'
    # í has a combining acute accent, ﬃ is a ligature
    bs = bistr('Hi\u0301\uFB03').casefold()
    assert bs.original == 'Hi\u0301\uFB03'
    assert bs.modified == 'hi\u0301ffi'
    assert bs.modified == bs.original.casefold()

    assert bs[:3].original == 'Hi\u0301'
    assert bs[:3].modified == 'hi\u0301'

    assert bs[4:5].original == '\uFB03'
    assert bs[4:5].modified == 'f'


def test_lower():
    bs = bistr('DİYARBAKIR').lower('en_US')
    assert bs.original == 'DİYARBAKIR'
    assert bs.modified == 'di̇yarbakir'

    bs = bistr('DİYARBAKIR').lower('tr_TR')
    assert bs.original == 'DİYARBAKIR'
    assert bs.modified == 'diyarbakır'


def test_upper():
    bs = bistr('straße').upper('de_DE')
    assert bs.original == 'straße'
    assert bs.modified == 'STRASSE'
    assert bs[4:6].original == 'ß'
    assert bs[4:6].modified == 'SS'

    bs = bistr('Diyarbakır').upper('tr_TR')
    assert bs.original == 'Diyarbakır'
    assert bs.modified == 'DİYARBAKIR'

def test_title():
    bs = bistr('istanbul').title('en_US')
    assert bs.original == 'istanbul'
    assert bs.modified == 'Istanbul'

    bs = bistr('istanbul').title('tr_TR')
    assert bs.original == 'istanbul'
    assert bs.modified == 'İstanbul'


def test_capitalize():
    bs = bistr('hello WORLD').capitalize('en_US')
    assert bs.original == 'hello WORLD'
    assert bs.modified == 'Hello world'
    assert bs.alignment == Alignment.identity(11)

    bs = bistr('τελικός').capitalize('el_GR')
    assert bs.original == 'τελικός'
    assert bs.modified == 'Τελικός'
    assert bs.alignment == Alignment.identity(7)

    bs = bistr('ἴΣ').capitalize('el_GR')
    assert bs.original == 'ἴΣ'
    assert bs.modified == 'Ἴς'
    assert bs.alignment == Alignment.identity(2)


def test_swapcase():
    bs = bistr('hello WORLD').swapcase('en_US')
    assert bs.original == 'hello WORLD'
    assert bs.modified == 'HELLO world'
    assert bs.alignment == Alignment.identity(11)

    # Ligatures/digraphs in title case don't have a swapped form
    bs = bistr('ǈepòta').swapcase('hr_HR')
    assert bs.original == 'ǈepòta'
    assert bs.modified == 'ǈEPÒTA'
    assert bs.alignment == Alignment.identity(6)

    bs = bistr('ǈepòta').normalize('NFKC').swapcase('hr_HR')
    assert bs.original == 'ǈepòta'
    assert bs.modified == 'lJEPÒTA'
    assert bs[0:2] == bistr('ǈ', 'lJ')


def test_normalize():
    # "Héllö" -- é is composed but ö has a combining diaeresis
    bs = bistr('H\u00E9llo\u0308').normalize('NFC')
    assert bs.original == 'H\u00E9llo\u0308'
    assert bs.modified == 'H\u00E9ll\u00F6'
    assert bs.modified == unicodedata.normalize('NFC', bs.original)
    assert bs[1:2] == bistr('\u00E9')
    assert bs[4:5] == bistr('o\u0308', '\u00F6')

    bs = bistr('H\u00E9llo\u0308').normalize('NFD')
    assert bs.original == 'H\u00E9llo\u0308'
    assert bs.modified == 'He\u0301llo\u0308'
    assert bs.modified == unicodedata.normalize('NFD', bs.original)
    assert bs[1:3] == bistr('\u00E9', 'e\u0301')
    assert bs[5:7] == bistr('o\u0308')


def test_readme():
    bs = bistr('𝕿𝖍𝖊 𝖖𝖚𝖎𝖈𝖐, 𝖇𝖗𝖔𝖜𝖓 🦊 𝖏𝖚𝖒𝖕𝖘 𝖔𝖛𝖊𝖗 𝖙𝖍𝖊 𝖑𝖆𝖟𝖞 🐶')
    bs = bs.normalize('NFKD')
    bs = bs.casefold()
    bs = bs.replace('🦊', 'fox')
    bs = bs.replace('🐶', 'dog')
    bs = bs.sub(r'[^\w\s]+', '')
    bs = bs[:19]
    assert bs.modified == 'the quick brown fox'
    assert bs.original == '𝕿𝖍𝖊 𝖖𝖚𝖎𝖈𝖐, 𝖇𝖗𝖔𝖜𝖓 🦊'


def test_equality():
    bs1 = bistr('  Hello world  ').strip().casefold()
    bs2 = bistr('  Hello world  ', 'hello world', Alignment([
        (0, 0),
        (2, 0),
        (3, 1),
        (4, 2),
        (5, 3),
        (6, 4),
        (7, 5),
        (8, 6),
        (9, 7),
        (10, 8),
        (11, 9),
        (12, 10),
        (13, 11),
        (15, 11),
    ]))
    assert bs1 == bs2


def test_alternative_regex():
    import regex

    bs = bistr('The quick, brown 🦊 jumps over the lazy 🐶')
    bs = bs.sub(regex.compile(r'\pS'), lambda m: unicodedata.name(m.group()))
    assert bs[17:25] == bistr('🦊', 'FOX FACE')
    assert bs[46:] == bistr('🐶', 'DOG FACE')



from typing import Iterator, Tuple, Union, overload


UString = Union[str, 'UnicodeString']


class BreakIterator:
    DONE: int

    @classmethod
    def createCharacterInstance(cls, locale: Locale) -> BreakIterator: ...

    @classmethod
    def createWordInstance(cls, locale: Locale) -> BreakIterator: ...

    @classmethod
    def createSentenceInstance(cls, locale: Locale) -> BreakIterator: ...

    def setText(self, text: UString) -> None: ...

    def first(self) -> int: ...

    def nextBoundary(self) -> int: ...

    def getRuleStatus(self) -> int: ...


class CaseMap:
    @classmethod
    def fold(cls, text: UString, edits: Edits) -> str: ...

    @classmethod
    def toLower(cls, locale: Locale, text: UString, edits: Edits) -> str: ...

    @classmethod
    def toUpper(cls, locale: Locale, text: UString, edits: Edits) -> str: ...

    @classmethod
    def toTitle(cls, locale: Locale, text: UString, edits: Edits) -> str: ...


class Edits:
    def __init__(self) -> None: ...

    def getFineIterator(self) -> Iterator[Tuple[bool, int, int, int, int, int]]: ...


class Locale:
    def __init__(self, name: str): ...


class Normalizer2:
    @classmethod
    def getNFCInstance(cls) -> Normalizer2: ...

    @classmethod
    def getNFDInstance(cls) -> Normalizer2: ...

    @classmethod
    def getNFKCInstance(cls) -> Normalizer2: ...

    @classmethod
    def getNFKDInstance(cls) -> Normalizer2: ...

    def normalize(self, text: UString) -> str: ...

    def hasBoundaryBefore(self, c: UString) -> bool: ...


class UnicodeString:
    def __init__(self, string: str): ...

    @overload
    def __getitem__(self, index: int) -> str: ...

    @overload
    def __getitem__(self, index: slice) -> UnicodeString: ...

    def __len__(self) -> int: ...

    @overload
    def countChar32(self) -> int: ...

    @overload
    def countChar32(self, start: int, length: int) -> int: ...

    def charAt(self, index: int) -> int: ...

    def char32At(self, index: int) -> int: ...



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from bistring import bistr, Token, Tokenization, Tokenizer
import pytest


def test_tokenization():
    text = bistr('  The quick, brown fox jumps over the lazy dog  ')
    text = text.replace(',', '')

    tokens = Tokenization(text, [
        Token.slice(text, 2, 5),
        Token.slice(text, 6, 11),
        Token.slice(text, 12, 17),
        Token.slice(text, 18, 21),
        Token.slice(text, 22, 27),
        Token.slice(text, 28, 32),
        Token.slice(text, 33, 36),
        Token.slice(text, 37, 41),
        Token.slice(text, 42, 45),
    ])
    assert tokens.text == text
    assert tokens.text_bounds(1, 3) == (6, 17)
    assert tokens.original_bounds(1, 3) == (6, 18)
    assert tokens.bounds_for_text(0, 13) == (0, 3)
    assert tokens.bounds_for_original(0, 13) == (0, 2)
    assert tokens.slice_by_text(36, 47).text == bistr('lazy dog')
    assert tokens.slice_by_original(36, 48).text == bistr('the lazy dog')
    assert tokens.snap_text_bounds(1, 13) == (2, 17)
    assert tokens.snap_original_bounds(36, 47) == (34, 46)

    tokens = tokens[1:-1]
    assert tokens.text.original == 'quick, brown fox jumps over the lazy'
    assert tokens.text.modified == 'quick brown fox jumps over the lazy'
    assert tokens.text_bounds(1, 3) == (6, 15)
    assert tokens.original_bounds(1, 3) == (7, 16)
    assert tokens.bounds_for_text(8, 14) == (1, 3)
    assert tokens.bounds_for_original(9, 15) == (1, 3)
    assert tokens.slice_by_text(8, 14).text == bistr('brown fox')
    assert tokens.slice_by_original(9, 15).text == bistr('brown fox')
    assert tokens.snap_text_bounds(8, 14) == (6, 15)
    assert tokens.snap_original_bounds(9, 15) == (7, 16)


def test_infer():
    text = 'the quick, brown fox'
    tokens = Tokenization.infer(text, ['the', 'quick', 'brown', 'fox'])
    assert tokens.substring(1, 3) == bistr('quick, brown')

    pytest.raises(ValueError, Tokenization.infer, text, ['the', 'quick', 'red', 'fox'])


def test_regex_tokenizer():
    from bistring import RegexTokenizer

    text = bistr(' 𝕿𝖍𝖊 𝖖𝖚𝖎𝖈𝖐, 𝖇𝖗𝖔𝖜𝖓 𝖋𝖔𝖝 𝖏𝖚𝖒𝖕𝖘 𝖔𝖛𝖊𝖗 𝖙𝖍𝖊 𝖑𝖆𝖟𝖞 𝖉𝖔𝖌 ')
    text = text.normalize('NFKD')
    text = text.casefold()

    tokenizer = RegexTokenizer(r'\w+')
    assert isinstance(tokenizer, Tokenizer)

    tokens = tokenizer.tokenize(text)
    assert tokens.text == text
    assert len(tokens) == 9
    assert tokens.text_bounds(0, 2) == (1, 10)
    assert tokens[0:2].text == text[1:10]
    assert len(tokens.slice_by_text(5, 10)) == 1
    assert len(tokens.slice_by_text(5, 11)) == 1
    assert len(tokens.slice_by_text(3, 13)) == 3


def test_splitting_tokenizer():
    from bistring import SplittingTokenizer

    text = bistr(' 𝕿𝖍𝖊 𝖖𝖚𝖎𝖈𝖐, 𝖇𝖗𝖔𝖜𝖓 𝖋𝖔𝖝 𝖏𝖚𝖒𝖕𝖘 𝖔𝖛𝖊𝖗 𝖙𝖍𝖊 𝖑𝖆𝖟𝖞 𝖉𝖔𝖌 ')
    text = text.normalize('NFKD')
    text = text.casefold()

    tokenizer = SplittingTokenizer(r'\s+')
    assert isinstance(tokenizer, Tokenizer)

    tokens = tokenizer.tokenize(text)
    assert tokens.text == text
    assert len(tokens) == 9
    assert tokens.text_bounds(0, 2) == (1, 11)
    assert tokens[0:2].text == text[1:11]
    assert len(tokens.slice_by_text(5, 10)) == 1
    assert len(tokens.slice_by_text(5, 11)) == 1
    assert len(tokens.slice_by_text(3, 13)) == 3


def test_character_tokenizer():
    from bistring import CharacterTokenizer

    text = bistr(' 𝕿𝖍𝖊 𝖖𝖚𝖎𝖈𝖐, 𝖇𝖗𝖔𝖜𝖓 𝖋𝖔𝖝 𝖏𝖚𝖒𝖕𝖘 𝖔𝖛𝖊𝖗 𝖙𝖍𝖊 𝖑𝖆𝖟𝖞 𝖉𝖔𝖌 ')

    tokenizer = CharacterTokenizer('en_US')
    assert isinstance(tokenizer, Tokenizer)

    tokens = tokenizer.tokenize(text)
    assert tokens.text == text
    assert all(token.text == text[i:i+1] for i, token in enumerate(tokens))


def test_word_tokenizer():
    from bistring import WordTokenizer

    text = bistr(' 𝕿𝖍𝖊 𝖖𝖚𝖎𝖈𝖐, 𝖇𝖗𝖔𝖜𝖓 𝖋𝖔𝖝 𝖏𝖚𝖒𝖕𝖘 𝖔𝖛𝖊𝖗 𝖙𝖍𝖊 𝖑𝖆𝖟𝖞 𝖉𝖔𝖌 ')

    tokenizer = WordTokenizer('en_US')
    assert isinstance(tokenizer, Tokenizer)

    tokens = tokenizer.tokenize(text)
    assert tokens.text == text
    assert len(tokens) == 9
    assert tokens.text_bounds(0, 2) == (1, 10)
    assert tokens[0:2].text == text[1:10]
    assert len(tokens.slice_by_text(5, 10)) == 1
    assert len(tokens.slice_by_text(5, 11)) == 1
    assert len(tokens.slice_by_text(3, 13)) == 3


def test_sentence_tokenizer():
    from bistring import SentenceTokenizer

    text = bistr('The following sentence is true.  The preceeding sentence, surprisingly, is false.')

    tokenizer = SentenceTokenizer('en_US')
    assert isinstance(tokenizer, Tokenizer)

    tokens = tokenizer.tokenize(text)
    assert tokens.text == text
    assert len(tokens) == 2
    assert tokens[0].text == text[:33]
    assert tokens[1].text == text[33:]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from bistring import Alignment
import pytest


def test_empty():
    pytest.raises(ValueError, Alignment, [])

    alignment = Alignment.identity(0)
    assert list(alignment) == [(0, 0)]

    assert alignment.original_bounds() == (0, 0)
    assert alignment.modified_bounds() == (0, 0)

    assert alignment.original_bounds(0, 0) == (0, 0)
    assert alignment.modified_bounds(0, 0) == (0, 0)


def test_identity():
    alignment = Alignment.identity(1, 16)

    assert alignment == Alignment((i, i) for i in range(1, 17))
    assert list(alignment) == [(i, i) for i in range(1, 17)]

    assert alignment.original_bounds() == (1, 16)
    assert alignment.modified_bounds() == (1, 16)

    assert alignment.original_bounds(4, 7) == (4, 7)
    assert alignment.modified_bounds(4, 7) == (4, 7)


def test_aligning():
    alignment = Alignment([(0, 0), (1, 2), (2, 4), (3, 6)])

    assert alignment.original_bounds() == (0, 3)
    assert alignment.modified_bounds() == (0, 6)

    assert alignment.original_bounds(0, 0) == (0, 0)
    assert alignment.original_bounds(0, 1) == (0, 1)
    assert alignment.original_bounds(0, 2) == (0, 1)
    assert alignment.original_bounds(0, 3) == (0, 2)
    assert alignment.original_bounds(1, 1) == (0, 1)
    assert alignment.original_bounds(1, 3) == (0, 2)
    assert alignment.original_bounds(1, 4) == (0, 2)
    assert alignment.original_bounds(2, 2) == (1, 1)
    assert alignment.original_bounds(2, 4) == (1, 2)
    assert alignment.original_bounds(2, 5) == (1, 3)
    assert alignment.original_bounds(3, 3) == (1, 2)

    assert alignment.modified_bounds(0, 0) == (0, 0)
    assert alignment.modified_bounds(0, 1) == (0, 2)
    assert alignment.modified_bounds(0, 2) == (0, 4)
    assert alignment.modified_bounds(0, 3) == (0, 6)
    assert alignment.modified_bounds(1, 1) == (2, 2)
    assert alignment.modified_bounds(2, 2) == (4, 4)


def test_canonicalization():
    assert Alignment([(0, 0), (1, 2), (1, 2), (2, 4)]) == Alignment([(0, 0), (1, 2), (2, 4)])

    assert Alignment([(0, 0), (1, 2)]) + Alignment([(1, 2), (2, 4)]) == Alignment([(0, 0), (1, 2), (2, 4)])


def _test_composition(first, second):
    composed = first.compose(second)

    of, ol = composed.original_bounds()
    mf, ml = composed.modified_bounds()

    assert (of, ol) == first.original_bounds()
    assert (mf, ml) == second.modified_bounds()

    for i in range(of, ol + 1):
        for j in range(i, ol + 1):
            assert composed.modified_bounds(i, j) == second.modified_bounds(first.modified_bounds(i, j))

    for i in range(mf, ml + 1):
        for j in range(i, ml + 1):
            assert composed.original_bounds(i, j) == first.original_bounds(second.original_bounds(i, j))


def test_compose():
    first = Alignment((i, 2 * i) for i in range(4))
    second = Alignment((i, 2 * i) for i in range(7))
    _test_composition(first, second)


def _test_identity_composition(alignment):
    _test_composition(alignment, Alignment.identity(alignment.modified_range()))
    _test_composition(Alignment.identity(alignment.original_range()), alignment)


def test_compose_identity():
    alignment = Alignment([
        (0, 2),
        (2, 2),
        (4, 4),
        (6, 6),
        (8, 6),
    ])

    # Modified sequence is smaller
    _test_identity_composition(alignment)

    # Original sequence is smaller
    _test_identity_composition(alignment.inverse())


def test_infer():
    assert Alignment.infer('test', 'test') == Alignment.identity(4)
    assert Alignment.infer('asdf', 'jkl;') == Alignment.identity(4)

    assert Alignment.infer('color', 'colour') == Alignment([
        (0, 0),
        (1, 1),
        (2, 2),
        (3, 3),
        (4, 4),
        (4, 5),
        (5, 6),
    ])

    assert Alignment.infer('color', 'colour') == Alignment.infer('colour', 'color').inverse()

    assert Alignment.infer("ab---", "ab") == Alignment([
        (0, 0),
        (1, 1),
        (2, 2),
        (3, 2),
        (4, 2),
        (5, 2),
    ])



#!/usr/bin/env python3

# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from pathlib import Path
from setuptools import setup


def readme():
    with open(Path(__file__).parent/'README.rst') as f:
        return f.read()


setup(
    name='bistring',
    version='0.3.0',
    description='Bidirectionally transformed strings',
    long_description=readme(),
    long_description_content_type='text/x-rst',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Text Processing :: General',
        'Typing :: Typed',
    ],
    keywords='bistring string non-destructive',
    url='https://github.com/microsoft/bistring',
    author='Microsoft Research Montreal',
    author_email='msrmtle@microsoft.com',
    license='MIT',
    packages=[
        'bistring',
    ],
    package_data={
        'bistring': [
            'py.typed',
        ],
    },
    zip_safe=False,
    test_suite='tests',
    setup_requires=[
        'pytest-runner',
    ],
    install_requires=[
        'pyicu',
    ],
    extras_require={
        'test': [
            'lxml',
            'mypy',
            'pytest',
            'regex',
        ],
    },
    tests_require=[
        'bistring[test]',
    ],
)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional
import unicodedata

from ._alignment import Alignment
from ._bistr import bistr
from ._token import CharacterTokenizer


@dataclass(frozen=True)
class AugmentedChar:
    """
    A single character (grapheme cluster) augmented with extra information.
    """

    top_category: str
    """
    The top-level Unicode category of the char (L, P, Z, etc.).
    """

    category: str
    """
    The specific Unicode category of the char (Lu, Po, Zs, etc.).
    """

    root: str
    """
    The root code point of the grapheme cluster.
    """

    folded: str
    """
    The case-folded form of the char.
    """

    normalized: str
    """
    The Unicode compatibility normalized form of the char.
    """

    original: str
    """
    The original form of the char.
    """

    @classmethod
    def cost_fn(cls, a: Optional[AugmentedChar], b: Optional[AugmentedChar]) -> int:
        """
        The cost function between augmented chars.  Each attribute contributes one "point" towards their distance.
        """

        if a is None or b is None:
            # cost(insert) + cost(delete) (4 + 4) should be more than cost(substitute) (6)
            return 4

        result = 0
        result += int(a.top_category != b.top_category)
        result += int(a.category != b.category)
        result += int(a.root != b.root)
        result += int(a.folded != b.folded)
        result += int(a.normalized != b.normalized)
        result += int(a.original != b.original)
        return result


TOKENIZER = CharacterTokenizer('root')


@dataclass(frozen=True)
class AugmentedString:
    """
    A string augmented with extra information about each character.
    """

    original: str
    """
    The original string.
    """

    chars: List[AugmentedChar]
    """
    The augmented characters of the string.
    """

    alignment: Alignment
    """
    The alignment between the original string and the augmented chars.
    """

    @classmethod
    def augment(cls, original: str) -> AugmentedString:
        normalized = bistr(original).normalize('NFKD')
        folded = bistr(normalized.modified).casefold()
        glyphs = TOKENIZER.tokenize(folded)

        chars = []
        for glyph in glyphs:
            fold_c = glyph.text.modified
            root = fold_c[0]

            norm_slice = folded.alignment.original_slice(glyph.start, glyph.end)
            norm_c = folded.original[norm_slice]

            orig_slice = normalized.alignment.original_slice(norm_slice)
            orig_c = normalized.original[orig_slice]

            cat = unicodedata.category(root)
            top_cat = cat[0]

            chars.append(AugmentedChar(top_cat, cat, root, fold_c, norm_c, orig_c))

        alignment = normalized.alignment
        alignment = alignment.compose(folded.alignment)
        alignment = alignment.compose(glyphs.alignment)
        return cls(original, chars, alignment)


def heuristic_infer(original: str, modified: str) -> bistr:
    """
    Infer the alignment between two strings with a "smart" heuristic.

    We use Unicode normalization and case folding to minimize differences that are due to case, accents, ligatures, etc.
    """

    aug_orig = AugmentedString.augment(original)
    aug_mod = AugmentedString.augment(modified)

    alignment = Alignment.infer(aug_orig.chars, aug_mod.chars, AugmentedChar.cost_fn)
    alignment = aug_orig.alignment.compose(alignment)
    alignment = alignment.compose(aug_mod.alignment.inverse())

    return bistr(original, modified, alignment)
